#!/usr/bin/python

#This program tries to locate the light-beams falling on the hawaii sensor.
#It locates the weighted center of the "star-matter", and then prints the coordinates.
#The input is a 1024 x 1024 pixel 32-bit (monochrome) tiff file, usually from cam2tiff.

#TODO: should we infer WxH from the image dimensions, rather than hard-coding them? If so, note
#that we still want to output the Hawaii coordinates, not the subarray ones.

#Copyright 2011 Richard Neill and Edmund Jackson. This is Free Software Licensed under GNU GPL v3+.

#INSTALLATION (see Readme.txt for more details):
#1. Install packages  python-numpy python-scipy ipython python-imaging python-matplotlib
#2. Fix the Python PIL Tiff definitions if necessary. Edit /usr/lib/python2.6/dist-packages/PIL/TiffImagePlugin.py
#   as per http://www.mailinglistarchive.com/html/image-sig@python.org/2011-04/msg00009.html

#How it works:
# * Find the brightest spot in the image, if it's 'bright' then
# * Look at all its neighbours, if they are bright then add them to the 'star matter'
# * Look at all the neighbours of the newly add star matter
# * Repeat until no new stars are added
# * Take the intensity weighted position of the star matter and call it a star
# * Blank out all the positions considered and go to the top
# * Adjust LUM_THRESH till it works.
# * Possible enhancement: examine the local statistics and find areas of high correlation. Thus even if the star is dim, if its blurry, we might be able to pull it out of the noise.



#FIXME :
#  * not guaranteed to terminate.  (try blurry beams with a 0 threshold)
#  * threshold is too sensitive. 0.18 works.
#  * fails on beams.tiff.
#  * Major (fatal) bug if reading directly from tiff file: Python treats 0xffffff as if it were a signed integer!
print "Warning, this code fails to read 32-bit tiffs, as required. It treats the unsigned numbers as signed!"
print "Do not use this; use the C-implementation, beam_finder.c instead."
exit(1)


import matplotlib.pyplot as pyplot
import numpy as numpy
import matplotlib.image as pimg
import array, sys, os, Image, ctypes
from functools import partial

def showhelp():
	basename = os.path.basename(sys.argv[0])
	print "This searches for light-beams and prints their co-ordinates (and some extra info.)."
	print "It also plots the beams that it identifies (NOT the original input file)"
	print "Usage:    " + basename + " [OPTS] infile.tiff    #tiff image, size "+ str(SIDE) + "x" + str(SIDE) + " px."
	print "Options:  -h, --help       #show help"
	print "Options:  -q, --quiet      #quiet: don't display a plot."
	print ""
	return


# ------------------------------/ Constants \----------------------------
# Define the the square dimension of the Hawaii array, and the image  (note that pixel-coords are 0-based)
SIDE = 1024

# Luminosity above which a pixel is considered part of a star.
# (tunable parameter, 0.1-0.2 seems to work, don't make it too low, avoid zero!)
LUM_THRESH = 0.1

# Max beams. Stop after finding at least this many (even if there might be more to find).
MAX_BEAMS = 10

# -------------------------/ Find the stars \----------------------------
# This is the first technique, super trival, just find the 4 largest valued
# pixels and return their locations

# Turn the index of a 1D flattened image into X of the 2D
def index_to_x (i):
    return int(numpy.floor(i / SIDE))

# Turn the index of a 1D flattened image into Y of the 2D
def index_to_y (i):
    return int(i - index_to_x(i) * SIDE)

# Return 1d index from [X, Y].  Used as lists are not Hashable
def xy_to_index (xy):
    return xy[1] + xy[0]*SIDE


# Return [X,Y] for the max pixel in an image
def max_2d (image):
    max_i = image[:,:,1].flatten().argmax()
    print "max2d:  " + str (index_to_x(max_i))  + " " + str(index_to_y(max_i))

    return [index_to_x(max_i), index_to_y(max_i)]

# Helper function for is_within_image. Is number representative of either an x or y coordinate?
def is_in_range (x):
	return (x >= 0) and (x < SIDE)

# Detecting each edge case on construction in torus is miserable. It is easier to
# detect and filter over/underflows. This accepts a vector of points and removes elements that over/underflow.
def within_image (data):
	return filter(lambda x: is_in_range(x[0]) and is_in_range(x[1]), data)

# Construct a torus around a point.
# NB the torus must fit entirely inside the image, or we will exceed the array-bounds and crash; hence the use
# of within_image(). This means we can't detect a beam that is right on the edge of the image.
def torus (pt):
	x = pt[0]
	y = pt[1]
	return within_image([[x-1, y-1],
	  [x, y-1],
	  [x+1, y-1],
	  [x-1, y],
	  [x+1, y],
	  [x-1, y+1],
	  [x, y+1],
	  [x+1, y+1]])

#  Get around idiotic lack of pair deref
def image_val (image, pt):
    return image[pt[0], pt[1]][1]

#  Have we observed this point yet ?
def unobserved (observed, pt):
    return not(xy_to_index(pt) in observed)

#  Bright enough to consider?
def is_bright_enough (image, thresh, pt):
    return image_val(image, pt) > thresh

def expand_about_point(x, image, observed, thresh):
    # Bright points we haven't yet seen
    new = filter(partial(unobserved, observed),
                 filter(partial(is_bright_enough, image, thresh),
                        torus(x)))

    # Update what we've seen
    map(observed.add, map(xy_to_index, torus(x)))

    return [observed, new]

# This is awful. Clean it up !!!!
def build_cluster(pt, image, thresh):
    observed = set([xy_to_index(pt)])
    star_matter = [pt]
    new = [pt]
    while (len(new) > 0):
        new_new = []
        for x in new:
            observed, n = expand_about_point(x, image, observed, thresh)
            new_new = new_new + n
            star_matter = star_matter + n
        new = new_new
    return [observed, star_matter]

# Infer the weighted centre of the star_matter and call in the centre of the star
def star_statistics (image, star_matter):
    # X
    x_v = sum(map(lambda x: x[1]*image_val(image, x), star_matter))    #NB is this right: 1,0 rather than 0,1 for x,y?

    # Y
    y_v = sum(map(lambda x: x[0]*image_val(image, x), star_matter))

    # Total Luminosity
    l = sum(map(lambda x: image_val(image, x), star_matter))

    # Number of pixels
    return {'w_y': y_v/l, 'w_x': x_v/l, 'l': l, 'n':len(star_matter)}

# Set all observed points to red (invisible to algo)
def blank_point(image, i):
    image[index_to_x(i), index_to_y(i)] = numpy.array([ 1.,  0.,  0.])

def blank_star(image, observed):
    map(partial(blank_point, image), observed)
    return image

def locate_star (image, thresh):
    observed, matter = build_cluster(max_2d(image), image, thresh)
    return [star_statistics(image, matter), observed]

def locate_all_stars(image, thresh):
    inferred_stars = []
    inferred_centre_lum = []
    while (len(inferred_stars) < MAX_BEAMS) and (image_val(image, max_2d(image)) > thresh):
	inferred_centre_lum += [image_val(image, max_2d(image))]
        s, o = locate_star(image, thresh)
        inferred_stars += [s]
        image = blank_star(image, o)
        #print "# found: " + str(len(inferred_stars))
    return [inferred_stars, inferred_centre_lum]

def print_star(star, lum):
    print "------------------"
    print 'centroid x: ' + str(star['w_x'])
    print 'centroid y: ' + str(star['w_y'])
    print 'pixel x: ' + str(round(star['w_x']))
    print 'pixel y: ' + str(round(star['w_y']))
    print 'total luminosity: ' + str(star['l'])
    print 'luminosity focus: ' + str(lum/star['l']) + "%"
    print 'total pixels: ' + str(star['n'])
    print 'average luminosity: ' + str(star['l']/star['n'])
    return

def print_output(stars, lums):
    print "The following are detected stars: "
    map(print_star, stars, lums)
    return

def tabulate_star(star, lum):
	print str(star['w_x']) +"\t"+str(star['w_y'])+"\t"+ str(round(star['w_x'])) +"\t"+ str(round(star['w_y']))+"\t"+ str(star['l'])+"\t"+ str(lum/star['l'])+"\t"+ str(star['n'])+"\t\t"+ str(star['l']/star['n'])
	return

#Convert X11 coordinates to Hawaii coordinates. Use `cam2tiff -h` for a diagram.
def x11_to_hawaii (X11_x,X11_y):
	if ((X11_x < SIDE/2) and (X11_y < SIDE/2)):
		H_q = "Q2"
		H_x = X11_x
		H_y = (SIDE/2) - 1 - X11_y
	elif ((X11_x < SIDE/2) and (X11_y >= SIDE/2)):
		H_q = "Q1"
		H_x = X11_x
		H_y = (SIDE) - 1 - X11_y
	elif ((X11_x >= SIDE/2) and (X11_y < SIDE/2)):
		H_q = "Q3"
		H_x = X11_x - (SIDE/2)
		H_y = (SIDE/2) - 1 - X11_y
	else:
		H_q = "Q0"
		H_x = X11_x - (SIDE/2)
		H_y = (SIDE) - 1 - X11_y
	return [H_q, H_x, H_y]

def tabulate_star2(star, lum):
	hawaii = x11_to_hawaii (round(star['w_x']), round(star['w_y']))
	print "%s	%d	%d	    %d		%5.4f	    %5.4f	%5.4f		%6.2f		%6.2f" % (
	hawaii[0],		#hawaii quad
	hawaii[1],		#hawaii X
	hawaii[2],		#hawaii Y
	star['n'],		#total pixels
	star['l'],		#total luminosity
	(lum/star['l']),	#central focus
	(star['l']/star['n']),	#avg luminosity
	star['w_x'],		#centroid_x  (in X11 coordinate frame)
	star['w_y']		#centroid_y
	)
	return

def tabulate_stars(stars, lums):
	#print "#centroid_x\tcentroid_y\tpix_x\tpix_y\ttotal_lum\tcentral_focus\ttotal_pixels\taverage_lum"
	#map(tabulate_star, stars, lums)
	print "#Found " + str(len(stars)) + " candidate beams, in descending luminosity:"
	print "#Hawaii_Coordinates	    Beam	Beam	    Central	Average		X11_Coordinates (topleft)"
	print "#QUAD	X	Y	    Pixels	Luminosity  Focus	Luminosity	Centroid Position (X,Y)"
	map(tabulate_star2, stars, lums)
	return

#-----------------------------------

# Command-line args.  (If this gets more complex, use getopt)
quiet = False
if (len(sys.argv) == 1):
	print "Wrong arguments. Use -h for help."
	exit (1)
elif (len(sys.argv) == 2):
	if ( (sys.argv[1] == "-h") or (sys.argv[1] == "--help") ):
		showhelp()
		exit (0)
	elif ( (sys.argv[1] == "-q") or (sys.argv[1] == "--quiet") ):
		print "Wrong arguments. Use -h for help."
		exit (1)
	else:
		filename = sys.argv[1]
elif (len(sys.argv) == 3):
	if ( (sys.argv[1] == "-q") or (sys.argv[1] == "--quiet") ):
		quiet = True
		filename = sys.argv[2]
	else:
		print "Wrong arguments. Use -h for help."
		exit (1)
else:
	print "Wrong arguments. Use -h for help."
	exit (1)


# Open the file to check that it exists.
try:
	FH = open(filename, 'rb')
except IOError:
	print "Error: cannot read input file: " + filename
	exit (1)

# Open the image. We want to work on 32-bit unsigned TIFF images, so a trivial patch to PIL is required:
# see http://www.mailinglistarchive.com/html/image-sig@python.org/2011-04/msg00009.html
try:
	im = Image.open(filename, 'r')
except  IOError, reason:
	print "Error: cannot process input file: " + filename   + " (reason: '" + str(reason) + "')."
	print "[NB: 'cannot identify image file' means that TiffImagePlugin.py must be (trivially) patched to support 32bit TIFF; see README.txt]"
	exit (1)

# Check that the image is a TIFF file of the correct size. Mode I is 32-bit integer pixels
if ( (im.format != "TIFF") or (im.size != (SIDE, SIDE)) or (im.mode != "I") ):
	print "Error: image in wrong format. Expect TIFF, " + str(SIDE) + "x" + str(SIDE) + " in mode 1. Found " + im.format + " " + str(im.size) + " mode " + im.mode
	exit (1)

# Convert image to numpy array (of ints).
#pix = im.load()

#image_in =  numpy.array (ctypes.c_uint32(im.getdata()) ).reshape(im.size[1], im.size[0])


# Python is totally brainless, and assumes (despite the obvious nature of images) that this is a signed integer!!!

raw = im.getdata()
raw2 = numpy.empty ( len(raw), dtype=int )
for i in range (0, len(raw)):
	raw2[i] = raw[i]
	if (raw2[i] < 0):
		raw2[i] += 4294967296
image_in = raw2.reshape(SIDE, SIDE)

#print pix [812, 924]
print image_in[812 ,924]
exit (1);


#Old way, starts by using stream to convert to raw image, then read raw data.
# Create the image from a tiff using ImageMagick (must be built to support 32-bit depths --with-quantum-depth=32)
#   /usr/local/bin/stream -map i -storage-type quantum -depth 32 test.tiff test.raw

# Read into array.
#raw = array.array('L')   # Unsigned 32bit integer
#raw.read(FH, SIDE*SIDE)
#FH.close()

# Normalise into floats [0.0, 1.0] so that we can display
#image_in = numpy.array(raw).repeat(3).reshape(SIDE, SIDE, 3)/4293844224.0

# Find the stars
stars, centre_lums = locate_all_stars(image_in, LUM_THRESH)
#print_output (stars, centre_lums)	#multiple lines
tabulate_stars(stars, centre_lums)	#tabulated

# Look at it
if (quiet != True):
	sys.stderr.write ("Plotting the beams that were identified.\n")
	img = pyplot.imshow(image_in)
	pyplot.show()		#on some older versions, need 'pyplot.show(img)'
