/* Convert camera data format into a 4 column data file of ascii values. Invoke with -h for help and explanation. */
/* This is based upon cam2tiff. Unlike cam2tiff, we don't left-shift the data to fill the MSBytes. */
/* We assume that the .cam file is in native Intel format, i.e. 32-bit little-endian, with MSByte zeroed. */

/* Copyright (C) Richard Neill 2011-2012, <cam2tiff at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>

#define eprintf(...)	fprintf(stderr, __VA_ARGS__)				/* Error printf: send to stderr  */
#define feprintf(...)	fprintf(stderr, __VA_ARGS__); exit (EXIT_FAILURE)	/* Fatal error printf: send to stderr and exit */

#define  CAM_EXTN  "cam"		/* Extensions: cam for infile; dat for outfile */
#define  DAT_EXTN  "dat"
#define  QUADRANTS 4			/* Number of columns in output == number of quadrants in chip */

typedef uint32_t sample_t;		/* 4 bytes per sample */

#define  CHUNKSIZE (sizeof(sample_t) * QUADRANTS)

void print_help() {
	eprintf ("cam2dat converts the raw data file from the IR camera into a 4-column ascii data file.\n"
		"Quadrants are de-interleaved into columns, but otherwise, the data remains interleaved: \n"
		"see 'cam2tiff -h' for a diagram of the raster pattern.\n"
		"\n"
		"The columns contain unsigned integers, from printf('%%u'), and represent quadrants 0-3.\n"
		"The file begins with some comments, prepended with '#'. The data is not otherwise\n"
		"transformed; unlike cam2tiff, there is no bitshift <<8.\n"
		"The file format is compatible with fftplot(1), and with Python's numpy.loadtxt().\n"
		"\n"
		"USAGE:    cam2dat [OPTIONS]    input.%s output.%s        #use '-' for stdout.\n"
		"OPTIONS:         -h, --help    show help\n"
		"\n",
		CAM_EXTN, DAT_EXTN);
}

int main ( int argc, char **argv ) {
 	int  opt; extern char *optarg; extern int optind, opterr, optopt;  			/* getopt */
	char *cam_filename, *dat_filename;
 	FILE *cam, *dat;
	struct stat stat_p;
	int number_of_samples;
	int i;
	sample_t buf[QUADRANTS];

	/* Get options and check for validity */
 	if ((argc > 1) && (!strcmp (argv[1], "--help"))) {	/* Support --help, without the full getopt_long */
		print_help();
 		exit (EXIT_SUCCESS);
 	}

 	while ((opt = getopt(argc, argv, "h")) != -1) {		/* Getopt */
 		switch (opt) {
 			case 'h':				/* help */
 				print_help();
 				exit(EXIT_SUCCESS);
 				break;
 			default:				/* Catches wrong options or options missing their arg */
 				feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], CAM_EXTN, DAT_EXTN );
 		}
 	}

 	if (argc - optind != 2){
 		feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], CAM_EXTN, DAT_EXTN );
	}

	/* Files. NB, can't use stdin, because we have to stat() it, and get the filesize to calculate number of samples (possible workaround: use argv[]) */
 	cam_filename = argv[optind];
 	dat_filename = argv[optind + 1];

 	/* Check cam_filename ends in .cam. */
 	if ( strcmp ( (cam_filename + strlen(cam_filename) - strlen(CAM_EXTN) -1), "."CAM_EXTN) ){
 		feprintf( "Error, input file %s is not a .%s file. (Wrong extension).\n", cam_filename, CAM_EXTN);
 	}

 	/* Open camera file (for reading) */
 	if ( ( cam = fopen ( cam_filename, "r" ) ) == NULL ) {
 		feprintf ( "Could not open %s for reading: %s\n", cam_filename, strerror(errno) );
 	}

 	if (stat (cam_filename, &stat_p) == -1){  					/* Get stats for file */
 		feprintf( "Error occurred attempting to stat %s\n", cam_filename);  	/* This error shouldn't happen, since we already managed to open the file! */
 	}

 	if (stat_p.st_size == 0){							/* Check it isn't empty. */
 		feprintf( "Error, input file %s is empty.\n", cam_filename);
 	}

	if (!strcmp(dat_filename, "-")){ /* Allow '-' for stdout */
		dat = stdout;
	}else{
		/* Check dat_filename ends in .dat. Useful to prevent overwriting the wrong file by accident. */
		if ( strcmp ( (dat_filename + strlen(dat_filename) - strlen(DAT_EXTN) -1), "."DAT_EXTN) ){
			feprintf("Error, output file %s is not a .%s file. (Wrong extension).\n", dat_filename, DAT_EXTN);
		}
		/* Open output file (for writing) */
		if ( ( dat = fopen ( dat_filename, "w" ) ) == NULL ) {
			feprintf ("Could not open %s for writing: %s\n", dat_filename, strerror(errno) );
		}
	}

 	/* Get file size in bytes (already known not to be zero). Then check it is a multiple of number of columns * sample-size, and convert to number of samples */
	number_of_samples = stat_p.st_size;
	if (number_of_samples % CHUNKSIZE != 0){
	 	feprintf ("Error: the input file is %d bytes long, which is not exactly divisible into %d columns of %d-byte samples.\n", number_of_samples, QUADRANTS, (int)sizeof(sample_t) );
 	}
	number_of_samples = number_of_samples / CHUNKSIZE;

	/* Create the header */
	fprintf (dat, "# This is the Hawaii .cam file, extracted by cam2dat into separate columns.\n");
	fprintf (dat, "# Each column remains raster-interleaved (see 'cam2tiff -h').\n");
	fprintf (dat, "# The columns are quadrants 0, 1, 2, 3 (left to right).\n");
	fprintf (dat, "# original_cam_file:\t%s\n", cam_filename);
	fprintf (dat, "# quadrants:\t%d\n",  QUADRANTS);
	fprintf (dat, "# samples:\t%d\n",  number_of_samples);

	/* Iterate over the input file, formatting and re-printing the output */
	for (i = 0; i < number_of_samples; i++){
		if ( fread ( buf, 1, CHUNKSIZE, cam ) != CHUNKSIZE ) { /* Read in data (expected in intel-native format) */
			fprintf ( stderr, "Error reading data from file: %s\n", strerror ( errno ) );
			exit(EXIT_FAILURE);
		}
		fprintf (dat, "%u\t%u\t%u\t%u\n", buf[0], buf[1], buf[2], buf[3]);		/* write out data, in ascii format, column 0 on left */
		//fprintf (dat, "0x%x\t0x%x\t0x%x\t0x%x\n", buf[0], buf[1], buf[2], buf[3]);	/* hex format - probably clearer.  But python's numpy.loadtxt() dislikes it. */
	}

	fclose ( cam );
	fclose ( dat );

	return (0);
}
