/* Convert camera data format into a 4 column data file of ascii values. Invoke with -h for help and explanation. */
/* This is the inverse of cam2dat. Mostly as a 2-step process for "dat2tiff" */
/* We output the .cam file in native Intel format, i.e. 32-bit little-endian, with MSByte zeroed. */

/* Copyright (C) Richard Neill 2012, <cam2tiff at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>

#define eprintf(...)	fprintf(stderr, __VA_ARGS__)				/* Error printf: send to stderr  */
#define feprintf(...)	fprintf(stderr, __VA_ARGS__); exit (EXIT_FAILURE)	/* Fatal error printf: send to stderr and exit */

#define  CAM_EXTN  "cam"		/* Extensions: cam for outfil; dat for infile */
#define  DAT_EXTN  "dat"
#define  QUADRANTS 4			/* Number of columns in input == number of quadrants in chip */
#define  LINE_MAXLEN 1024
#define  MAX_24BIT  0xffffff		/* 24-bit */

typedef uint32_t sample_t;		/* 4 bytes per sample */

void print_help() {
	eprintf ("dat2cam inverts cam2dat.\n"
		"It reads a 4-column ascii data file, and converts it back to the raw data file for the IR camera.\n"
		"Columns are interleaved into quadrants, but the data ordering is otherwise unchanged: \n"
		"\n"
		"The input file columns usually contain unsigned integers, from printf('%%u'), and represent quadrants 0-3.\n"
		"If input format is floating point, it is scaled and offset, to map [-fullscale, fullscale] => [0, 0xffffff]\n"
		"Comments in the input file, prepended with '#', are discarded.\n"
		"\n"
		"USAGE:    dat2cam [OPTIONS]    input.%s output.%s     #use '-' for stdin,stdout.\n"
		"OPTIONS:    -f  fullscale      input is in floating point format, [+/- fullscale].\n"
		"            -h, --help         show help\n"
		"\n",
		DAT_EXTN, CAM_EXTN);
}

int main ( int argc, char **argv ) {
 	int  opt; extern char *optarg; extern int optind, opterr, optopt;  			/* getopt */
	char *cam_filename, *dat_filename;
 	FILE *cam, *dat;
	struct stat stat_p;
	double fullscale = 0, scalefactor=0, x, max_24bit=(float)MAX_24BIT;
	int column, line=1, is_float=0;
	sample_t data[QUADRANTS]={0,0,0,0};
	char buffer[LINE_MAXLEN];   	/* 1024 */
	char *token, *end="";		/* pointer to string token */
	char *whitespace = "\t\n ";	/* space, newline or tab */

	/* Get options and check for validity */
 	if ((argc > 1) && (!strcmp (argv[1], "--help"))) {	/* Support --help, without the full getopt_long */
		print_help();
 		exit (EXIT_SUCCESS);
 	}

 	while ((opt = getopt(argc, argv, "hf:")) != -1) {		/* Getopt */
 		switch (opt) {
 			case 'h':				/* help */
 				print_help();
 				exit(EXIT_SUCCESS);
 				break;
			case 'f':
				fullscale = atof (optarg);
				is_float = 1;
				if (fullscale == 0){
					feprintf ("fullscale can't be zero!\n");
				}else{
					scalefactor = max_24bit / (2 * fullscale);
				}
				break;
 			default:				/* Catches wrong options or options missing their arg */
 				feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], DAT_EXTN, CAM_EXTN );
 		}
 	}

 	if (argc - optind != 2){
 		feprintf ( "Syntax: %s [OPTIONS] input.%s output.%s.   #use -h for help\n", argv[0], DAT_EXTN, CAM_EXTN );
	}

 	dat_filename = argv[optind];
 	cam_filename = argv[optind + 1];


	/* If dat filename is "-", use stdin. Otherwise check dat_filename ends in .dat. */
	if (!strcmp (dat_filename, "-")){
		dat = stdin;
	}else{
		if ( strcmp ( (dat_filename + strlen(dat_filename) - strlen(DAT_EXTN) -1), "."DAT_EXTN) ){
			feprintf( "Error, input file %s is not a .%s file. (Wrong extension).\n", dat_filename, DAT_EXTN);
		}
		if ( ( dat = fopen ( dat_filename, "r" ) ) == NULL ) {				/* Open camera file (for reading) */
			feprintf ( "Could not open %s for reading: %s\n", dat_filename, strerror(errno) );
		}
		if (stat (dat_filename, &stat_p) == -1){  					/* Get stats for file */
			feprintf( "Error occurred attempting to stat %s\n", dat_filename);  	/* This error shouldn't happen, since we already managed to open the file! */
		}
		if (stat_p.st_size == 0){							/* Check it isn't empty. */
			feprintf( "Error, input file %s is empty.\n", dat_filename);
		}
	}

	if (!strcmp(cam_filename, "-")){
		cam = stdout;
	}else{
		/* Check cam_filename ends in .cam Useful to prevent overwriting the wrong file by accident. */
		if ( strcmp ( (cam_filename + strlen(cam_filename) - strlen(CAM_EXTN) -1), "."CAM_EXTN) ){
			feprintf("Error, output file %s is not a .%s file. (Wrong extension).\n", cam_filename, CAM_EXTN);
		}
		/* Open output file (for writing) */
		if ( ( cam = fopen ( cam_filename, "w" ) ) == NULL ) {
			feprintf ("Could not open %s for writing: %s\n", cam_filename, strerror(errno) );
		}
	}

	/* Read each line of source into buffer */
	while (fgets(buffer, LINE_MAXLEN, dat) != NULL){

		token = strtok (buffer, whitespace);	/* Find the first whitespace-delimited token within the string. If the string is an empty line, nothing will be found. */
		column = 0;
		while (token != NULL){
			if (token[0] == '#') {    /* Discard comments */
				break;
			}

			/* Parse the data. strtol() converts the string (which may be decimal or hex) to a long. Catch anything illegal. */
			if (column < QUADRANTS){
				if (is_float){
					x = scalefactor * strtod(token,&end);  /* floating point input, scaled and shifted */
					if (abs(x) > max_24bit/2){
						feprintf ("Value %s is too large for range [0, 0x%x], with scale-factor specified as %f.\n", token, MAX_24BIT, fullscale);
					}
					data[column] = (unsigned int)round( x + (max_24bit / 2));
				}else{
					data[column] = strtoul(token,&end,0);  /* unsigned integer data */
				}
			}else{
				feprintf ("Error in source at line %d: too many columns. Expect %d\n", line, QUADRANTS);
			}
			if (errno != 0 || *end != 0 || end == token){	/* Were there any unused chars in the string, eg trailing garbage. */
				feprintf("Error in source at line %d: strtoul() couldn't parse column %d, '%s'.\n", column, column, token);
			}

			token = strtok (NULL, whitespace); 	/* Find any subsequent whitespace-delimited token within the string */
			column++;
		}
 		if ((column > 0) && (column < QUADRANTS)){	/* Too few cols in this line! */
 			feprintf("Error in source at line %d: too few arguments. Expect 4, got %d\n", line, column);
 		}else if (column > 0){				/* Valid line */

			/* Write it out, in raw format */
			fwrite(data, sizeof(sample_t), QUADRANTS, cam);
		}
		line++;
	}

	fclose ( cam );
	fclose ( dat );

	return (0);
}
