/* Generate a test image for cam2tiff. Invoke with -h for help. */

/* Copyright (C) Richard Neill 2006-2011, <cam2tiff at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

/* Output in Intel Native 32-bit format, LSB, Mid, MSB, ZERO   (the upper byte is zero because we really only use 24-bit data) */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <asm/byteorder.h>

#define NUM_ROWS 512			/* Number of rows, per quadrant */
#define NUM_COLS 512			/* Number of columns, per quadrant */

#define SMALL_NUM_ROWS  48		/* For small image */
#define SMALL_NUM_COLS  64

#define BEAM_X_0 100			/* Beam positions within quad */
#define BEAM_Y_0 200
#define BEAM_X_1 250
#define BEAM_Y_1 300
#define BEAM_X_2 350
#define BEAM_Y_2 400
#define BEAM_X_3 450
#define BEAM_Y_3 500

uint32_t samples[4];			/* 4 bytes data */
FILE *outfile;

void printhelp(){
	fprintf (stderr, "gencamtest generates an image in cam format to test cam2tiff.\n"
			 "The test image, has 4 quadrants, each %dx%d pixels:\n"
			 "   * Bottom right (Q0)\n"
			 "   * Bottom left  (Q1)\n"
			 "   * Top left     (Q2)\n"
			 "   * Top right    (Q3)\n"
			 "See 'cam2tiff -h' for a diagram.\n"
			 "Each pixel is (intel-native) little-endian uint32, containing a 24-bit signal.\n"
			 "Thus the byte order is: Q0_L:M:H:0, Q1_L:M:H:0; Q2_L:M:H:0; Q3_L:M:H:0.\n"
			 "To view the result, use 'od -D file.cam', or convert with 'cam2tiff' and use 'display'.\n"
			 "\n"
			 "USAGE: gencamtest --type file.cam\n"
			 "    --white                 All pixels white: 0xFFFFFF.\n"
			 "    --black                 All pixels black: 0x000000.\n"
			 "    --striped               Alternate pixels black, white (vertical lines).\n"
			 "    --gradients             Q0 almost white (0xffeedd), Q1 vertical gradient (black at bottom),\n"
			 "                            Q2 almost black (0x000122), Q3 horizontal gradient (black at left).\n"
			 "    --4quad                 4 quadrants different colours:  Q0 white (0xffffff), Q1 black (0x000000)\n"
			 "                            Q2 dark (0x505050), Q3 light (0x909090).\n"
			 "    --4quaddark             Same as --4quad, but data in LSByte only.\n"
			 "    --align                 For alignment: large bright spots (Q0 darker...Q3 brighter), black background,\n"
			 "                            quad borders white. x-axis (fast) light-grey, y-axis (slow) dark-grey.\n"
			 "    --noise                 High level rand() noise. Quads have different brightnesses.\n"
			 "    --beams                 Simple, perfectly aligned, single-pixel bright beams; black background.\n"
			 "    --blurrybeams           Blurry, bright beams (3px radius); black background.\n"
			 "    --blurryskewbeams       Blurry, bright beams (Approx 3px radius, not very round); black background.\n"
			 "    --blurrybeamsnoise      Blurry, bright beams, imperfectly-round; faint noisy background. S/N good.\n"
			 "    --faintblurrybeamsnoise Fuzzy, faint beams, uniform faint noisy background. S/N poor.\n"
			 "    --smallgradients        Like --gradients, but outputting a small, %d x %d image.\n"
			 "\n"
			 "  Beam 0 (level 0xffffff): Quad 0 (X=%d, Y=%d), Cam.co-ords: (X=%d, Y=%d),  X11.co-ords: (X=%d, Y=%d)\n"
			 "  Beam 1 (level 0x0000ff): Quad 1 (X=%d, Y=%d), Cam.co-ords: (X=%d, Y=%d),  X11.co-ords: (X=%d, Y=%d)\n"
			 "  Beam 2 (level 0x505050): Quad 2 (X=%d, Y=%d), Cam.co-ords: (X=%d, Y=%d),  X11.co-ords: (X=%d, Y=%d)\n"
			 "  Beam 3 (level 0x909090): Quad 3 (X=%d, Y=%d), Cam.co-ords: (X=%d, Y=%d), X11.co-ords: (X=%d, Y=%d)\n"
			 "\n"
			 "  (CAM co-ordinates are wrt. the Hawaii sensor:  origin bottom-left, x = right, y = upwards.)\n"
			 "  (X11 co-ordinates are wrt. X11/display/stream: origin top-left,    x = right, y = downwards.)\n"
			 "\n"
			 "\n",
			 NUM_ROWS, NUM_COLS, SMALL_NUM_COLS, SMALL_NUM_ROWS,
			 BEAM_X_0, BEAM_Y_0, BEAM_X_0 + NUM_COLS,  BEAM_Y_0,             BEAM_X_0 + NUM_COLS,   NUM_ROWS*2 - BEAM_Y_0,
			 BEAM_X_1, BEAM_Y_1, BEAM_X_1,             BEAM_Y_1,             BEAM_X_1,              NUM_ROWS*2 - BEAM_Y_1,
			 BEAM_X_2, BEAM_Y_2, BEAM_X_2,             BEAM_Y_2 + NUM_ROWS,  BEAM_X_2,              NUM_ROWS   - BEAM_Y_2,
			 BEAM_X_3, BEAM_Y_3, BEAM_X_3 + NUM_COLS,  BEAM_Y_3 + NUM_ROWS,  BEAM_X_3 + NUM_COLS,   NUM_ROWS   - BEAM_Y_3
			);
}

void write_bytes(){
	samples[0] = __cpu_to_le32 (samples[0]);
	samples[1] = __cpu_to_le32 (samples[1]);
	samples[2] = __cpu_to_le32 (samples[2]);
	samples[3] = __cpu_to_le32 (samples[3]);
	if ( fwrite ( samples, 1, sizeof ( samples ), outfile )  != sizeof ( samples ) ) {
		fprintf ( stderr, "Error writing: %s\n",  strerror ( errno ) );
		exit ( 1 );
	}
}


int main ( int argc, char **argv ) {
	char *outfile_filename;
	char *type;
	unsigned int row;
	unsigned int col;
	srand (time (NULL));		/* Seed random number generator (omit this to make the "random" images reproducable). */

	if (argc < 2){
		printhelp();
		exit (1);
	}
	if ((!strcmp (argv[1], "-h")) || (!strcmp (argv[1], "--help"))){
		printhelp();
		exit (1);
	}
	if ( argc != 3 ) {
		fprintf ( stderr, "Syntax: %s --type test.cam   #-h for help.\n", argv[0] );
		exit ( 1 );
	}
	type = argv[1];
	outfile_filename = argv[2];

	if ( ( outfile = fopen ( outfile_filename, "w" ) ) == NULL ) {
		fprintf ( stderr, "Could not open %s for writing: %s\n",
			  outfile_filename, strerror ( errno ) );
		exit ( 1 );
	}

	/* Output different format, depending on parameter */

	if (!strcmp (type, "--white")){					/* All white pixels, FFFFFF */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0xffffff;
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--black")){				/* All pixels black: 000000 */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--striped")){				/* Alternate pixels black or white: vertical lines */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				if (col %2 == 0){
					samples[0] = samples[1] = samples[2] = samples[3] = 0xffffff;
				}else{
					samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;
				}
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--gradients")){			/* Almost black, almost white, 2 gradients */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] =  0xffeedd ; /* Bottom right: all white (not perfect white, as a check byte ordering) */
				samples[1] = row << 15 ; /* Bottom left: vertical gradient (black at bottom) */
				samples[2] =  0x000122 ; /* Top left: all black (again, not perfectly) */
				samples[3] = col << 15 ; /* Top right: horizontal gradient (black at left) */
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--smallgradients")){			/* --gradients, but much smaller. */

		for ( row = 0 ; row < SMALL_NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < SMALL_NUM_COLS ; col++ ) {
				samples[0] =  0xffeedd ; /* Bottom right: all white (not perfect white, as a check byte ordering) */
				samples[1] = row << 18 ; /* Bottom left: vertical gradient (black at bottom) */
				samples[2] =  0x000122 ; /* Top left: all black (again, not perfectly) */
				samples[3] = col << 18 ; /* Top right: horizontal gradient (black at left) */
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--4quad")){			/*  4 quadrants different colours:  BR white, BL black, TL dark, TR light */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = 0xffffff;
				samples[1] = 0x000000;
				samples[2] = 0x505050;
				samples[3] = 0x909090;
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--4quaddark")){		/* 4 quadrants, different colours, all almost black, data in LSByte only. */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = 0x0000ff;
				samples[1] = 0x000000;
				samples[2] = 0x000050;
				samples[3] = 0x000090;
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--align")){			/* Simple, bright spots, 10px radius. show quad borders, black background. */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;	/* black background */
				if ( ( pow((abs(col - BEAM_X_0)),2) + pow((abs(row - BEAM_Y_0)),2)) < 100 ){   /* round spots, 10px radius. */
					samples[0] = 0xffffff;
				}
				if ( ( pow((abs(col - BEAM_X_1)),2) + pow((abs(row - BEAM_Y_1)),2)) < 100 ){
					samples[1] = 0x333333;
				}
				if ( ( pow((abs(col - BEAM_X_2)),2) + pow((abs(row - BEAM_Y_2)),2)) < 100 ){
					samples[2] = 0x777777;
				}
				if ( ( pow((abs(col - BEAM_X_3)),2) + pow((abs(row - BEAM_Y_3)),2)) < 100 ){
					samples[3] = 0xaaaaaa;
				}
				if ( (col == 0) || (row == 0) ){				/* quad borders white */
					samples[0] = samples[1] = samples[2] = samples[3] = 0xffffff;
				}
				if ( (row == 20) && (col > 10) && (col < NUM_COLS - 20) ){	/* x-axis light grey */
					samples[0] = samples[1] = samples[2] = samples[3] = 0xbbbbbb;
				}
				if ( (col == 20) && (row > 10) && (row < NUM_ROWS - 20) ){	/* y-axis dark grey */
					samples[0] = samples[1] = samples[2] = samples[3] = 0x666666;
				}
				write_bytes ();
			}
		}

 	}else if (!strcmp (type, "--noise")){			/* High level noise; quads are slightly different brightnesses) */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = rand() / (RAND_MAX / 0xffffff);   /* to scale with integer maths, divide by reciprocal */
				samples[1] = rand() / (RAND_MAX / 0xff);
				samples[2] = rand() / (RAND_MAX / 0x505050);
				samples[3] = rand() / (RAND_MAX / 0x909090);
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--beams")){			/* Simple, perfectly aligned, bright beams. black background. */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;
				if (col == BEAM_X_0 && row == BEAM_Y_0){
					samples[0] = 0xffffff;
				}
				if (col == BEAM_X_1 && row == BEAM_Y_1){
					samples[1] = 0x0000ff;
				}
				if (col == BEAM_X_2 && row == BEAM_Y_2){
					samples[2] = 0x505050;
				}
				if (col == BEAM_X_3 && row == BEAM_Y_3){
					samples[3] = 0x909090;
				}
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--blurrybeams")){			/* Blurry, bright beams. black background. */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {		/* Center pixel bright; surrounding 4: half brightness; surrounding 8: quarter brightness */
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;
				if (col == BEAM_X_0 && row == BEAM_Y_0){
					samples[0] = 0xffffff;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 1){
					samples[0] = 0xffffff / 2;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 2){
					samples[0] = 0xffffff / 4;
				}
				if (col == BEAM_X_1 && row == BEAM_Y_1){
					samples[1] = 0x0000ff;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 1){
					samples[1] = 0x0000ff / 2;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 2){
					samples[1] = 0x0000ff / 4;
				}
				if (col == BEAM_X_2 && row == BEAM_Y_2){
					samples[2] = 0x505050;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 1){
					samples[2] = 0x505050 / 2;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 2){
					samples[2] = 0x505050 / 4;
				}
				if (col == BEAM_X_3 && row == BEAM_Y_3){
					samples[3] = 0x909090;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 1){
					samples[3] = 0x909090 / 2;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 2){
					samples[3] = 0x909090 / 4;
				}
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--blurryskewbeams")){			/* Blurry, bright beams, not very round or centered, black background. */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {		/* Center pixel bright; surrounding 4: approx half brightness; surrounding 8: approx quarter brightness */
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = samples[1] = samples[2] = samples[3] = 0x000000;
				if (col == BEAM_X_0 && row == BEAM_Y_0){
					samples[0] = 0xffffff;
				}else if (col == BEAM_X_0 && row == BEAM_Y_0 + 1){  //skew
					samples[0] = 0xffffff;
				}else if (col == BEAM_X_0 && row == BEAM_Y_0 + 2){  //skew
					samples[0] = 0xffffff;
				}else if (col == BEAM_X_0 && row == BEAM_Y_0 + 3){  //skew
					samples[0] = 0xffffff;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 1){
					samples[0] = 0xffffff / 2;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 2){
					samples[0] = 0xffffff / 4;
				}
				if (col == BEAM_X_1 && row == BEAM_Y_1){
					samples[1] = 0x0000ff;
				}else if (col == BEAM_X_1 + 1 && row == BEAM_Y_1){  //skew
					samples[1] = 0x0000ff;
				}else if (col == BEAM_X_1 + 2 && row == BEAM_Y_1){  //skew
					samples[1] = 0x0000ff;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 1){
					samples[1] = 0x0000ff / 2;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 2){
					samples[1] = 0x0000ff / 4;
				}
				if (col == BEAM_X_2 && row == BEAM_Y_2){
					samples[2] = 0x505050;
				}else if ((col == BEAM_X_2 + 1) && (row == BEAM_Y_2 + 1)){  //skew
					samples[2] = 0x505050;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 1){
					samples[2] = 0x505050 / 4;			    //reduced
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 2){
					samples[2] = 0x505050 / 8;			    //reduced
				}
				if (col == BEAM_X_3 && row == BEAM_Y_3){
					samples[3] = 0x909090;
				}else if ((col == BEAM_X_3 + 1) && (row == BEAM_Y_3 + 1)){  //skew
					samples[3] = 0x909090;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 1){
					samples[3] = 0x909090 / 16;			    //reduced
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 2){
					samples[3] = 0x909090 / 32;			    //reduced
				}
				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--blurrybeamsnoise")){		/* Fuzzy, faint beams (non-centered), uniform, faint noisy background.  */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = rand() / (RAND_MAX / 0x08ffff);	  /* Noise at the 18-bit level */
				samples[1] = rand() / (RAND_MAX / 0x08ffff);
				samples[2] = rand() / (RAND_MAX / 0x08ffff);
				samples[3] = rand() / (RAND_MAX / 0x08ffff);

				if (col == BEAM_X_0 && row == BEAM_Y_0){	 /* Brightest beam at the 24-bit level */
					samples[0] += 0xffffff;			 /* <-- this centre pixel will overflow the 24-bit level, and appear dark, unless we rotate right later */
				}else if ((col - BEAM_X_0) == 1 && row == BEAM_Y_0){
					samples[0] += 0xffffff / 3;		//off-center
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 1){
					samples[0] += 0xffffff / 2;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 2){
					samples[0] += 0xffffff / 4;
				}
				if (col == BEAM_X_1 && row == BEAM_Y_1){
					samples[1] += 0x0000ff;
				}else if (col == BEAM_X_1 && (row - BEAM_Y_1) == 1){
					samples[0] += 0x0000ff / 1;		//off-center
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 1){
					samples[1] += 0x0000ff / 2;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 2){
					samples[1] += 0x0000ff / 4;
				}
				if (col == BEAM_X_2 && row == BEAM_Y_2){
					samples[2] += 0x505050;
				}else if (((int)col - BEAM_X_2) == -1 && row == BEAM_Y_2){
					samples[0] += 0x505050 / 6;		//off-center
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 1){
					samples[2] += 0x505050 / 2;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 2){
					samples[2] += 0x505050 / 4;
				}
				if (col == BEAM_X_3 && row == BEAM_Y_3){
					samples[3] += 0x909090;
				}else if (col == BEAM_X_3 && ((int)col - BEAM_Y_3) == -1){
					samples[0] += 0x909090 / 1;		//off-center
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 1){
					samples[3] += 0x909090 / 2;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 2){
					samples[3] += 0x909090 / 4;
				}
				samples[0] >>= 1;				/* Attenute everything to the 23-bit level */
				samples[1] >>= 1;				/* Might be useful to attenuate more, but then not human-viewable. */
				samples[2] >>= 1;				/* Must rotate at least 1 bit: see above addition/overflow */
				samples[3] >>= 1;

				write_bytes ();
			}
		}

	}else if (!strcmp (type, "--faintblurrybeamsnoise")){		/* Fuzzy, faint beams, uniform, faint noisy background. (very poor S/N ratio)  */

		for ( row = 0 ; row < NUM_ROWS ; row++ ) {
			for ( col = 0 ; col < NUM_COLS ; col++ ) {
				samples[0] = rand() / (RAND_MAX / 0x4fffff);	  /* Noise at the 22-bit level */
				samples[1] = rand() / (RAND_MAX / 0x4fffff);
				samples[2] = rand() / (RAND_MAX / 0x4fffff);
				samples[3] = rand() / (RAND_MAX / 0x4fffff);

				if (col == BEAM_X_0 && row == BEAM_Y_0){	 /* Brightest beam at the 24-bit level */
					samples[0] += 0xffffff;			 /* <-- this centre pixel will overflow the 24-bit level, and appear dark, unless we rotate right later */
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 1){
					samples[0] += 0xffffff / 2;
				}else if ((abs(col - BEAM_X_0) + abs(row - BEAM_Y_0)) == 2){
					samples[0] += 0xffffff / 4;
				}
				if (col == BEAM_X_1 && row == BEAM_Y_1){
					samples[1] += 0x0000ff;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 1){
					samples[1] += 0x0000ff / 2;
				}else if ((abs(col - BEAM_X_1) + abs(row - BEAM_Y_1)) == 2){
					samples[1] += 0x0000ff / 4;
				}
				if (col == BEAM_X_2 && row == BEAM_Y_2){
					samples[2] += 0x505050;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 1){
					samples[2] += 0x505050 / 2;
				}else if ((abs(col - BEAM_X_2) + abs(row - BEAM_Y_2)) == 2){
					samples[2] += 0x505050 / 4;
				}
				if (col == BEAM_X_3 && row == BEAM_Y_3){
					samples[3] += 0x909090;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 1){
					samples[3] += 0x909090 / 2;
				}else if ((abs(col - BEAM_X_3) + abs(row - BEAM_Y_3)) == 2){
					samples[3] += 0x909090 / 4;
				}
				samples[0] >>= 1;				/* Attenute everything to the 23-bit level */
				samples[1] >>= 1;				/* Might be useful to attenuate more, but then not human-viewable. */
				samples[2] >>= 1;				/* Must rotate at least 1 bit: see above addition/overflow */
				samples[3] >>= 1;

				write_bytes ();
			}
		}

	}else{
		fprintf (stderr, "Sorry, format %s not recognised. Use -h for list.\n", type);
	}


	fclose ( outfile );
	return 0;
}
